from cast_ai.se.constants import PROD_BUCKET, GCP_AUTH_CMD

from datetime import datetime, timezone, timedelta
from google.cloud.storage import Client
from google.cloud.exceptions import Forbidden

import subprocess
import json
import logging
import os


class SnapshotController:
    def __init__(self, cast_api_key: str, default_cluster_id: str, json_key_path: str):
        self._logger = logging.getLogger(__name__)
        self._api_key = cast_api_key
        self._cluster_id = default_cluster_id
        self._json_key_path = json_key_path
        self.raw_snapshot = None
        self._storage_client = None
        if os.path.isfile(self._json_key_path):
            self._storage_client = Client.from_service_account_json(self._json_key_path)
        self._get_raw_snapshot()

    def _is_json_key_auth_sufficient(self) -> bool:
        if self._storage_client:
            try:
                list(self._storage_client.list_blobs(bucket_or_name=PROD_BUCKET,
                                                     max_results=1,
                                                     prefix=self._cluster_id))
                return True
            except Forbidden:
                self._logger.warning("Authenticated using specified key file, but access denied.")
            except Exception as e:
                self._logger.error(f"Error trying to authenticate using json: {e}")
        else:
            return False

    def _authenticate_via_gcloud(self):
        try:
            subprocess.check_output(GCP_AUTH_CMD, text=True, shell=True)
            self._check_bucket_access()
            return
        except subprocess.CalledProcessError as e:
            self._logger.error(f"Error running gcloud auth application-default login: {e}")
        except Forbidden:
            self._logger.error("Authenticated using gcloud, but access denied.")
        except Exception as e:
            self._logger.critical(f"Error trying to authenticate via gcloud: {e}")
        raise RuntimeError("Error during gcloud authentication")

    def _already_authenticated(self) -> bool:
        try:
            self._check_bucket_access()
            return True
        except Forbidden:
            self._logger.error("Assumed Authenticated,...but access denied")
        except Exception as e:
            self._logger.critical(f"Error trying to list buckets when assuming already authenticated: {e}")
        return False

    def _check_bucket_access(self):
        self._storage_client = Client()
        list(self._storage_client.list_blobs(bucket_or_name=PROD_BUCKET,
                                             max_results=1,
                                             prefix=self._cluster_id))

    def _gcp_authenticate(self):
        if self._already_authenticated():
            return
        if not self._is_json_key_auth_sufficient():
            self._authenticate_via_gcloud()

    def _get_snapshot_prefix(self):
        current_utc_time = datetime.now(timezone.utc)
        previous_minute_time = current_utc_time - timedelta(minutes=1)
        formatted_previous_minute_time = previous_minute_time.strftime("%Y-%m-%dT%H:%M")
        prefix = f"{self._cluster_id}/{formatted_previous_minute_time}"
        return prefix

    def _get_raw_snapshot(self) -> None:
        self._gcp_authenticate()
        prefix = self._get_snapshot_prefix()
        blobs = list(self._storage_client.list_blobs(bucket_or_name=PROD_BUCKET, max_results=3, prefix=prefix))
        if not len(blobs):
            self._logger.error("No snapshots found for current time")
            raise RuntimeError("No snapshots found for current time")
        blob = blobs[-1]
        self.raw_snapshot = json.loads(blob.download_as_bytes().decode('utf-8'))



from typing import Dict, Any, List
import logging


class RefinedSnapshot:
    def __init__(self):
        self.pdbs = []
        self.workloads = WorkloadObject()
        self.karpenter_flag = False
        self.windows_flag = False


class WorkloadObject:
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self.deployments = []
        self.replica_sets = []
        self.daemon_sets = []
        self.stateful_sets = []
        self.jobs = []

    def get_workload_type(self, workload_type: str) -> List[Any]:
        try:
            return vars(self)[workload_type]
        except Exception as e:
            self._logger.error(f"Error returning internal sub-object of WorkloadObject workload_type={workload_type}")
            return []

    def add_item(self, item_type: str, item_data: Dict[str, Any]) -> None:
        match item_type:
            case 'deployments':
                item_data["index"] = len(self.deployments)
                self.deployments.append(item_data)
            case 'replica_sets':
                item_data["index"] = len(self.replica_sets)
                self.replica_sets.append(item_data)
            case 'daemon_sets':
                item_data["index"] = len(self.daemon_sets)
                self.daemon_sets.append(item_data)
            case 'stateful_sets':
                item_data["index"] = len(self.stateful_sets)
                self.stateful_sets.append(item_data)
            case 'jobs':
                item_data["index"] = len(self.jobs)
                self.jobs.append(item_data)
            case _:
                raise ValueError(f"Invalid item type: {item_type}")



import logging
from typing import Dict, Any

from cast_ai.se.models.misc_data_models import ReportDetailLevel
from cast_ai.se.services.snapshot_refinery_svc import SnapshotRefinery


class SnapshotAnalyzer:
    def __init__(self, raw_snapshot: Dict[str, Any], auto_refine: bool = True):
        self._logger = logging.getLogger(__name__)
        self._raw_snapshot = raw_snapshot
        self._refinery = SnapshotRefinery()
        if auto_refine:
            self.refine_snapshot()

    def refine_snapshot(self):
        self._refinery.refine_snapshot(self._raw_snapshot)

    def get_refined_snapshot_report(self, detail_lvl: ReportDetailLevel = ReportDetailLevel.BASIC):
        return self._refinery.reporter.generate_refined_snapshot_report(detail_lvl)

    def get_refined_snapshot_csv_report(self):
        self._refinery.reporter.generate_refined_snapshot_csv_report(cluster_id=self._raw_snapshot["clusterId"])



API_SERVER = "https://api.cast.ai"
EXTERNAL_CLUSTER_PREFIX = "/v1/kubernetes/external-clusters/"
AUDIT_PREFIX = "/v1/audit"
CLUSTER_Q = "?clusterId="
CLUSTERS_PREFIX = "/v1/kubernetes/clusters/"
POLICIES_POSTFIX = "/policies"
NODES_POSTFIX = "/nodes"
RECONCILE_POSTFIX = "/reconcile"

CAST_NS = "castai-agent"

DEMO_ON_CRONJOB = "hibernate-resume"
DEMO_OFF_CRONJOB = "hibernate-pause"

LOG_DIR = "logs"

REQUIRED_TOOLS = ["kubectl", "jq"]

WIN_GET_DEPLOYMENTS_CMD = ('get deployments -n default'
                           ' --output=jsonpath="{range .items[*]} {.metadata.name} {end}"')

LINUX_GET_DEPLOYMENTS_CMD = 'get deployments -n default --output=jsonpath="{.items[*].metadata.name}"'


LINUX_GET_NONZERO_DEPLOYMENTS_CMD = ('get deployments -n default -o json | jq -r ".items[] | '
                                     'select(.spec.replicas!=0) | .metadata.name"')

WIN_GET_NONZERO_DEPLOYMENTS_CMD = ('get deployments -n default'
                                   ' -o json | jq -r ".items[] | select(.spec.replicas!=0) | .metadata.name"')

GCP_AUTH_SCOPE = "https://www.googleapis.com/auth/cloud-platform"
GCP_AUTH_CMD = "gcloud auth application-default login"
PROD_BUCKET = "prod-master-console-cluster-snapshots-snapshotstore"
PROD_PROJECT = "prod-master-scl0"

POD_SPEC_KEYWORDS = ["nodeSelector",
                     "topologySpreadConstraints",
                     "affinity",
                     "runtimeClassName",
                     "tolerations"]

CSV_REPORT_HEADERS = ["Type",
                      "Name",
                      "Namespace",
                      "nodeSelector",
                      "topologySpreadConstraints",
                      "affinity",
                      "runtimeClassName",
                      "tolerations",
                      "No Requests"]

CSV_REPORT_HEADERS_PDB = ["Name",
                          "Namespace",
                          "Disruptions Allowed",
                          "Current Healthy"]


K8S_WORKLOAD_TYPES = ["deploymentList",
                      "replicaSetList",
                      "daemonSetList",
                      "statefulSetList",
                      "jobList"]

WORKLOAD_MAP = {'deploymentList': 'deployments',
                'replicaSetList': 'replica_sets',
                'daemonSetList': 'daemon_sets',
                'statefulSetList': 'stateful_sets',
                'jobList': 'jobs'}

WORKLOAD_TYPES = ['deployments', 'replica_sets', 'daemon_sets', 'stateful_sets', 'jobs']

NON_RELEVANT_NAMESPACES = ["castai-agent"]

CLOUD_TAINTS = ["node.kubernetes.io/unreachable",
                "node.kubernetes.io/not-ready",
                "node.kubernetes.io/unschedulable",
                "node.kubernetes.io/disk-pressure",
                "node.kubernetes.io/memory-pressure",
                "node.kubernetes.io/pid-pressure",
                "node.kubernetes.io/network-unavailable"]





class RefinedSnapshotAnalysis:
    def __init__(self):
        self.workloads_observed = {}
        self.pdbs_observed = {
            "namespaces": {}
        }
        self.total_counters = {
            "nodeSelector": 0,
            "topologySpreadConstraints": 0,
            "affinity": 0,
            "runtimeClassName": 0,
            "no_requests": 0,
            "tolerations": 0,
            "total_pdbs": 0,
            "total_workloads": 0
        }

    def __getitem__(self, key):
        return self.workloads_observed[key]
